@@ -9,6 +9,7 @@
 #include "include/media_kit_video/video_output.h"
 #include "include/media_kit_video/texture_gl.h"
 #include "include/media_kit_video/texture_sw.h"
+#include "include/media_kit_video/gl_render_thread.h"
 
 #include <epoxy/egl.h>
 #include <epoxy/glx.h>
@@ -18,7 +19,10 @@
 struct _VideoOutput {
   GObject parent_instance;
   TextureGL* texture_gl;
-  GdkGLContext* gdk_gl_context;
+  EGLDisplay egl_display; /* EGL display for mpv rendering (shared with flutter). */
+  EGLConfig egl_config;   /* EGL config from Flutter (for compatibility). */
+  EGLContext egl_context; /* Isolated EGL context (non-shared). */
+  EGLSurface egl_surface; /* Place holder surface for activating egl context */
   guint8* pixel_buffer;
   TextureSW* texture_sw;
   GMutex mutex; /* Only used in S/W rendering. */
@@ -30,6 +34,7 @@ struct _VideoOutput {
   TextureUpdateCallback texture_update_callback;
   gpointer texture_update_callback_context;
   FlTextureRegistrar* texture_registrar;
+  GLRenderThread* gl_render_thread;
   gboolean destroyed;
 };
 
@@ -49,26 +54,27 @@ static void video_output_dispose(GObject* object) {
     fl_texture_registrar_unregister_texture(self->texture_registrar,
                                             FL_TEXTURE(self->texture_gl));
     
-    // Make GL context current before freeing mpv_render_context
-    // mpv_render_context_free requires the GL context to be current
-    if (self->gdk_gl_context != NULL && self->render_context != NULL) {
-      GdkGLContext* current_ctx = gdk_gl_context_get_current();
-      gboolean need_restore = (current_ctx != self->gdk_gl_context);
-      if (need_restore) {
-        gdk_gl_context_make_current(self->gdk_gl_context);
-      }
-      mpv_render_context_free(self->render_context);
-      self->render_context = NULL;
-      if (need_restore) {
-        if (current_ctx != NULL) {
-          gdk_gl_context_make_current(current_ctx);
-        } else {
-          gdk_gl_context_clear_current();
+    // Clean up EGL resources in dedicated GL thread
+    if (self->render_context != NULL || self->egl_context != EGL_NO_CONTEXT) {
+      self->gl_render_thread->PostAndWait([self]() {
+        // Free mpv_render_context with our isolated EGL context
+        if (self->render_context != NULL) {
+          if (self->egl_context != EGL_NO_CONTEXT) {
+            eglMakeCurrent(self->egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, self->egl_context);
+          }
+          mpv_render_context_free(self->render_context);
+          self->render_context = NULL;
         }
-      }
+        
+        // Clean up EGL context
+        if (self->egl_context != EGL_NO_CONTEXT) {
+          eglDestroyContext(self->egl_display, self->egl_context);
+          self->egl_context = EGL_NO_CONTEXT;
+        }
+      });
     }
+    
     g_object_unref(self->texture_gl);
-    g_object_unref(self->gdk_gl_context);
   }
   // S/W
   if (self->texture_sw) {
@@ -83,8 +89,6 @@ static void video_output_dispose(GObject* object) {
   }
   
   g_mutex_clear(&self->mutex);
-  g_print("media_kit: VideoOutput: video_output_dispose: %ld\n",
-          (gint64)self->handle);
   G_OBJECT_CLASS(video_output_parent_class)->dispose(object);
 }
 
@@ -94,7 +98,10 @@ static void video_output_class_init(VideoOutputClass* klass) {
 
 static void video_output_init(VideoOutput* self) {
   self->texture_gl = NULL;
-  self->gdk_gl_context = NULL;
+  self->egl_display = EGL_NO_DISPLAY;
+  self->egl_config = NULL;
+  self->egl_context = EGL_NO_CONTEXT;
+  self->egl_surface = EGL_NO_SURFACE;
   self->texture_sw = NULL;
   self->pixel_buffer = NULL;
   self->handle = NULL;
@@ -105,6 +112,7 @@ static void video_output_init(VideoOutput* self) {
   self->texture_update_callback = NULL;
   self->texture_update_callback_context = NULL;
   self->texture_registrar = NULL;
+  self->gl_render_thread = NULL;
   self->destroyed = FALSE;
   g_mutex_init(&self->mutex);
 }
@@ -112,10 +120,11 @@ static void video_output_init(VideoOutput* self) {
 VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
                               FlView* view,
                               gint64 handle,
-                              VideoOutputConfiguration configuration) {
-  g_print("media_kit: VideoOutput: video_output_new: %ld\n", handle);
+                              VideoOutputConfiguration configuration,
+                              GLRenderThread* gl_render_thread) {
   VideoOutput* self = VIDEO_OUTPUT(g_object_new(video_output_get_type(), NULL));
   self->texture_registrar = texture_registrar;
+  self->gl_render_thread = gl_render_thread;
   self->handle = (mpv_handle*)handle;
   self->width = configuration.width;
   self->height = configuration.height;
@@ -127,45 +136,96 @@ VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
   }
   self->configuration.enable_hardware_acceleration = TRUE;
 #endif
-  mpv_set_option_string(self->handle, "video-sync", "audio");
-  mpv_set_option_string(self->handle, "video-timing-offset", "0");
+  
   gboolean hardware_acceleration_supported = FALSE;
+  
+  // Get EGL display and config in main thread (where Flutter context is available)
+  // Only attempt if hardware acceleration is enabled
   if (self->configuration.enable_hardware_acceleration) {
-    GError* error = NULL;
-    GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(view));
-    self->gdk_gl_context = gdk_window_create_gl_context(window, &error);
-    if (error == NULL) {
-      // OpenGL context must be made current before creating mpv render context.
-      gdk_gl_context_realize(self->gdk_gl_context, &error);
-      if (error == NULL) {
-        // Make the OpenGL context current.
-        gdk_gl_context_make_current(self->gdk_gl_context);
-        // Create |FlTextureGL| and register it.
-        self->texture_gl = texture_gl_new(self);
-        if (fl_texture_registrar_register_texture(
-                texture_registrar, FL_TEXTURE(self->texture_gl))) {
-          // Request H/W decoding.
+    EGLDisplay flutter_display = eglGetCurrentDisplay();
+    EGLContext flutter_context = eglGetCurrentContext();
+    
+    if (flutter_display != EGL_NO_DISPLAY && flutter_context != EGL_NO_CONTEXT) {
+      self->egl_display = flutter_display;
+      
+      // Get Flutter's EGL config by querying its context
+      EGLint config_id = 0;
+      if (eglQueryContext(flutter_display, flutter_context, EGL_CONFIG_ID, &config_id)) {
+        // Retrieve the actual EGLConfig from the config ID
+        EGLint num_configs = 0;
+        EGLint config_attribs[] = { EGL_CONFIG_ID, config_id, EGL_NONE };
+        
+        if (eglChooseConfig(flutter_display, config_attribs, &self->egl_config, 1, &num_configs) && num_configs > 0) {
+          g_print("media_kit: VideoOutput: Got Flutter's EGL display (%p) and config (ID: %d)\n", 
+                  flutter_display, config_id);
+          
+          // Create texture_gl in main thread (needed by mpv callback)
+          self->texture_gl = texture_gl_new(self);
+          if (!fl_texture_registrar_register_texture(
+                  texture_registrar, FL_TEXTURE(self->texture_gl))) {
+            g_printerr("media_kit: VideoOutput: Failed to register texture.\n");
+            g_object_unref(self->texture_gl);
+            self->texture_gl = NULL;
+            self->egl_config = NULL;
+          }
+        } else {
+          g_printerr("media_kit: VideoOutput: Failed to get Flutter's EGL config by ID.\n");
+          self->egl_config = NULL;
+        }
+      } else {
+        g_printerr("media_kit: VideoOutput: Failed to query Flutter's EGL config ID.\n");
+        self->egl_config = NULL;
+      }
+    } else {
+      g_printerr("media_kit: VideoOutput: Failed to get Flutter's EGL display or context.\n");
+    }
+  }
+  
+  // Initialize mpv in dedicated GL render thread
+  gl_render_thread->PostAndWait([self, &hardware_acceleration_supported]() {
+    mpv_set_option_string(self->handle, "video-sync", "audio");
+    // Causes frame drops with `pulse` audio output. (SlotSun/dart_simple_live#42)
+    // mpv_set_option_string(self->handle, "video-timing-offset", "0");
+    
+    if (self->texture_gl != NULL && 
+        self->egl_display != EGL_NO_DISPLAY && 
+        self->egl_config != NULL) {
+      
+      // Bind OpenGL ES API (Flutter uses OpenGL ES on Linux)
+      eglBindAPI(EGL_OPENGL_ES_API);
+      
+      // Create an isolated EGL context using Flutter's config
+      // Using the SAME egl_display and egl_config as Flutter for maximum compatibility
+      const EGLint context_attribs[] = {
+          EGL_CONTEXT_CLIENT_VERSION, 2,
+          EGL_NONE
+      };
+      
+      self->egl_context = eglCreateContext(self->egl_display, self->egl_config, 
+                                           EGL_NO_CONTEXT, context_attribs);
+      
+      if (self->egl_context != EGL_NO_CONTEXT) {
+        g_print("media_kit: VideoOutput: Created isolated EGL context: %p (display: %p, using Flutter's config)\n", 
+                self->egl_context, self->egl_display);
+        
+        // Make our isolated context current for initialization (surfaceless)
+        if (eglMakeCurrent(self->egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, self->egl_context)) {
+          // Initialize mpv with our isolated EGL context
           mpv_opengl_init_params gl_init_params{
               [](auto, auto name) {
-                GdkDisplay* display = gdk_display_get_default();
-                if (GDK_IS_WAYLAND_DISPLAY(display)) {
-                  return (void*)eglGetProcAddress(name);
-                }
-                if (GDK_IS_X11_DISPLAY(display)) {
-                  return (void*)glXGetProcAddressARB((const GLubyte*)name);
-                }
-                g_assert_not_reached();
-                return (void*)NULL;
+                return (void*)eglGetProcAddress(name);
               },
               NULL,
           };
+          
           mpv_render_param params[] = {
               {MPV_RENDER_PARAM_API_TYPE, (void*)MPV_RENDER_API_TYPE_OPENGL},
               {MPV_RENDER_PARAM_OPENGL_INIT_PARAMS, (void*)&gl_init_params},
               {MPV_RENDER_PARAM_INVALID, (void*)0},
               {MPV_RENDER_PARAM_INVALID, (void*)0},
           };
-          // VAAPI acceleration requires passing X11/Wayland display.
+          
+          // VAAPI acceleration requires passing X11/Wayland display
           GdkDisplay* display = gdk_display_get_default();
           if (GDK_IS_WAYLAND_DISPLAY(display)) {
             params[2].type = MPV_RENDER_PARAM_WL_DISPLAY;
@@ -174,8 +234,8 @@ VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
             params[2].type = MPV_RENDER_PARAM_X11_DISPLAY;
             params[2].data = gdk_x11_display_get_xdisplay(display);
           }
-          if (mpv_render_context_create(&self->render_context, self->handle,
-                                        params) == 0) {
+          
+          if (mpv_render_context_create(&self->render_context, self->handle, params) == 0) {
             mpv_render_context_set_update_callback(
                 self->render_context,
                 [](void* data) {
@@ -183,28 +243,43 @@ VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
                   if (self->destroyed) {
                     return;
                   }
-                  fl_texture_registrar_mark_texture_frame_available(
-                      self->texture_registrar, FL_TEXTURE(self->texture_gl));
+                  // Asynchronously notify render (don't block mpv thread)
+                  video_output_notify_render(self);
                 },
                 self);
             hardware_acceleration_supported = TRUE;
-            g_print("media_kit: VideoOutput: Using H/W rendering.\n");
+            g_print("media_kit: VideoOutput: H/W rendering with isolated EGL context in dedicated thread.\n");
+          } else {
+            g_printerr("media_kit: VideoOutput: Failed to create mpv_render_context.\n");
+            eglDestroyContext(self->egl_display, self->egl_context);
+            self->egl_context = EGL_NO_CONTEXT;
           }
+        } else {
+          g_printerr("media_kit: VideoOutput: Failed to make isolated EGL context current. Error: 0x%x\n", eglGetError());
+          eglDestroyContext(self->egl_display, self->egl_context);
+          self->egl_context = EGL_NO_CONTEXT;
         }
+      } else {
+        g_printerr("media_kit: VideoOutput: Failed to create isolated EGL context. Error: 0x%x\n", eglGetError());
       }
     }
-    if (error) {
-      g_print("media_kit: VideoOutput: GError: %d\n", error->code);
-      g_print("media_kit: VideoOutput: GError: %s\n", error->message);
-    }
+    // If hardware acceleration is not supported or disabled, fall back to software rendering
+  });
+  // hardware_acceleration_supported is already set by the lambda
+  
+  // If hardware acceleration failed and texture was created, clean it up
+  if (!hardware_acceleration_supported && self->texture_gl != NULL) {
+    fl_texture_registrar_unregister_texture(texture_registrar, 
+                                            FL_TEXTURE(self->texture_gl));
+    g_object_unref(self->texture_gl);
+    self->texture_gl = NULL;
   }
 #ifdef MPV_RENDER_API_TYPE_SW
   if (!hardware_acceleration_supported) {
-    // H/W rendering failed somewhere down the line. Fallback to S/W
-    // rendering.
+    g_printerr("media_kit: VideoOutput: S/W rendering.\n");
+    // H/W rendering failed. Fallback to S/W rendering.
     self->pixel_buffer = g_new0(guint8, SW_RENDERING_PIXEL_BUFFER_SIZE);
     self->texture_gl = NULL;
-    self->gdk_gl_context = NULL;
     self->texture_sw = texture_sw_new(self);
     if (fl_texture_registrar_register_texture(texture_registrar,
                                               FL_TEXTURE(self->texture_sw))) {
@@ -217,11 +292,6 @@ VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
         mpv_render_context_set_update_callback(
             self->render_context,
             [](void* data) {
-              // Usage on single-thread is not a concern with pixel buffers
-              // unlike OpenGL. So, I'd like to render on a separate thread
-              // for slowing the UI thread as little as possible. It's a pity
-              // that software rendering is feeling faster than hardware
-              // rendering due to fucked-up GTK.
               gdk_threads_add_idle(
                   [](gpointer data) -> gboolean {
                     VideoOutput* self = (VideoOutput*)data;
@@ -252,7 +322,6 @@ VideoOutput* video_output_new(FlTextureRegistrar* texture_registrar,
                   data);
             },
             self);
-        g_print("media_kit: VideoOutput: Using S/W rendering.\n");
       }
     }
   }
@@ -304,8 +373,20 @@ mpv_render_context* video_output_get_render_context(VideoOutput* self) {
   return self->render_context;
 }
 
-GdkGLContext* video_output_get_gdk_gl_context(VideoOutput* self) {
-  return self->gdk_gl_context;
+EGLDisplay video_output_get_egl_display(VideoOutput* self) {
+  return self->egl_display;
+}
+
+EGLContext video_output_get_egl_context(VideoOutput* self) {
+  return self->egl_context;
+}
+
+EGLSurface video_output_get_egl_surface(VideoOutput* self) {
+  return self->egl_surface;
+}
+
+GLRenderThread* video_output_get_gl_render_thread(VideoOutput* self) {
+  return self->gl_render_thread;
 }
 
 guint8* video_output_get_pixel_buffer(VideoOutput* self) {
@@ -434,3 +515,46 @@ void video_output_notify_texture_update(VideoOutput* self) {
     self->texture_update_callback(id, width, height, context);
   }
 }
+
+void video_output_notify_render(VideoOutput* self) {
+  if (self->destroyed || !self->gl_render_thread) {
+    return;
+  }
+  // Post combined check_and_resize + render task to GL thread (asynchronously)
+  self->gl_render_thread->Post([self]() {
+    video_output_check_and_resize(self);
+    video_output_render(self);
+  });
+}
+
+void video_output_check_and_resize(VideoOutput* self) {
+  if (self->destroyed || !self->texture_gl) {
+    return;
+  }
+  
+  TextureGL* texture = self->texture_gl;
+  gint64 required_width = video_output_get_width(self);
+  gint64 required_height = video_output_get_height(self);
+  
+  if (required_width < 1 || required_height < 1) {
+    return;
+  }
+  
+  // Check if resize is needed through texture_gl
+  texture_gl_check_and_resize(texture, required_width, required_height);
+}
+
+void video_output_render(VideoOutput* self) {
+  if (self->destroyed) {
+    return;
+  }
+  
+  // H/W rendering
+  if (self->texture_gl && self->render_context) {
+    texture_gl_render(self->texture_gl);
+    
+    // Notify Flutter that a new frame is available
+    fl_texture_registrar_mark_texture_frame_available(
+        self->texture_registrar, FL_TEXTURE(self->texture_gl));
+  }
+}
